import re # two problems

from ..protocol import Protocol, Variable, Struct, NUMERIC_TYPE_SIZES
from ..writer import Writer, TextUtil
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "C"


class CWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".h"

    def __init__(self, p: Protocol, extra_args: dict[str,any] = {}):
        super().__init__(protocol=p, tab="    ")

        self.embed_protocol = extra_args["embed_protocol"]

        self.type_mapping["byte"] = "uint8_t"
        self.type_mapping["bool"] = "bool"
        self.type_mapping["uint16"] = "uint16_t"
        self.type_mapping["int16"] = "int16_t"
        self.type_mapping["uint32"] = "uint32_t"
        self.type_mapping["int32"] = "int32_t"
        self.type_mapping["uint64"] = "uint64_t"
        self.type_mapping["int64"] = "int64_t"
        self.type_mapping["float"] = "float"
        self.type_mapping["double"] = "double"
        self.type_mapping["string"] = "char*"

        self.base_serializers: dict[str,str] = {
            "byte":   "UInt8",
            "bool":   "Bool",
            "uint16": "UInt16",
            "int16":  "Int16",
            "uint32": "UInt32",
            "int32":  "Int32",
            "uint64": "UInt64",
            "int64":  "Int64",
            "float":  "Float",
            "double": "Double",
        }

        self.base_defaults: dict[str,str] = {
            "byte":   "0",
            "bool":   "false",
            "uint16": "0",
            "int16":  "0",
            "uint32": "0",
            "int32":  "0",
            "uint64": "0",
            "int64":  "0",
            "float":  "0.0f",
            "double": "0.0",
        }

        self.subs: list[tuple[str,str]] = []
        self.prefix = "beschi_"
        if self.protocol.namespace != None:
            self.subs = [("beschi", self.protocol.namespace), ("BESCHI", self.protocol.namespace.upper())]
            self.prefix = f"{self.protocol.namespace}_"

    def err_check_return(self):
        self.write_line(f"if (err != {self.prefix.upper()}ERR_OK) {{")
        self.indent_level += 1
        self.write_line("return err;")
        self.indent_level -= 1
        self.write_line("}")

    def deserializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"err = {self.prefix}_Read{self.base_serializers[self.protocol.list_size_type]}(r, &({accessor}{var.name}_len));")
            self.err_check_return()
            if var.vartype in NUMERIC_TYPE_SIZES or var.vartype == "string":
                pref = ""
            else:
                pref = self.prefix
            self.write_line(f"{accessor}{var.name} = ({pref}{self.type_mapping[var.vartype]}*)malloc(sizeof({pref}{self.type_mapping[var.vartype]}) * {accessor}{var.name}_len);")
            self.write_line(f"if ({accessor}{var.name} == NULL) {{ return {self.prefix.upper()}ERR_ALLOCATION_FAILURE; }}")
            if var.vartype == "string":
                self.write_line(f"{accessor}{var.name}_els_len = ({self.get_native_string_size()}*)malloc(sizeof({self.get_native_string_size()}) * {accessor}{var.name}_len);")
                self.write_line(f"if ({accessor}{var.name} == NULL) {{ return {self.prefix.upper()}ERR_ALLOCATION_FAILURE; }}")
            idx = self.indent_level
            self.write_line(f"for ({self.get_native_list_size()} i{idx} = 0; i{idx} < {accessor}{var.name}_len; i{idx}++) {{")
            self.indent_level += 1
            inner = Variable(self.protocol, f"{var.name}[i{idx}]", var.vartype)
            self.deserializer(inner, accessor)
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            idx_search = re.match(r"(.*)\[(i\d+)\]$", var.name)
            if idx_search != None:
                name = idx_search.group(1)
                idx = idx_search.group(2)
                self.write_line(f"err = {self.prefix}_ReadString(r, &({accessor}{var.name}), &({accessor}{name}_els_len[{idx}]));")
            else:
                self.write_line(f"err = {self.prefix}_ReadString(r, &({accessor}{var.name}), &({accessor}{var.name}_len));")
        elif var.vartype in NUMERIC_TYPE_SIZES:
            self.write_line(f"err = {self.prefix}_Read{self.base_serializers[var.vartype]}(r, &({accessor}{var.name}));")
        else:
            self.write_line(f"err = {self.prefix}{var.vartype}_FromBytes(r, &({accessor}{var.name}));")
        self.err_check_return()

    def serializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"err = {self.prefix}_Write{self.base_serializers[self.protocol.list_size_type]}(w, &({accessor}{var.name}_len));")
            self.err_check_return()
            idx = self.indent_level
            self.write_line(f"for ({self.get_native_list_size()} i{idx} = 0; i{idx} < {accessor}{var.name}_len; i{idx}++) {{")
            self.indent_level += 1
            inner = Variable(self.protocol, f"{var.name}[i{idx}]", var.vartype)
            self.serializer(inner, accessor)
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            idx_search = re.match(r"(.*)\[(i\d+)\]$", var.name)
            if idx_search != None:
                name = idx_search.group(1)
                idx = idx_search.group(2)
                self.write_line(f"err = {self.prefix}_WriteString(w, &({accessor}{var.name}), &({accessor}{name}_els_len[{idx}]));")
            else:
                self.write_line(f"err = {self.prefix}_WriteString(w, &({accessor}{var.name}), &({accessor}{var.name}_len));")
        elif var.vartype in NUMERIC_TYPE_SIZES:
            self.write_line(f"err = {self.prefix}_Write{self.base_serializers[var.vartype]}(w, &({accessor}{var.name}));")
        else:
            self.write_line(f"err = {self.prefix}{var.vartype}_WriteBytes(w, &({accessor}{var.name}));")
        self.err_check_return()

    def gen_measurement(self, st: Struct, accessor: str = "") -> tuple[list[str], int]:
        lines: list[str] = []
        accum = 0

        if st.is_simple():
            lines.append(f"*size = {self.protocol.get_size_of(st.name)};")
        else:
            size_init = "*size = 0;"
            lines.append(size_init)

            for var in st.members:
                if var.is_list:
                    accum += NUMERIC_TYPE_SIZES[self.protocol.list_size_type]
                    idx = self.indent_level
                    if var.is_simple(True):
                        lines.append(f"*size += {accessor}{var.name}_len * {self.protocol.get_size_of(var.vartype)};")
                    elif var.vartype == "string":
                        lines.append(f"for ({self.get_native_list_size()} i{idx} = 0; i{idx} < {accessor}{var.name}_len; i{idx}++) {{")
                        lines.append(f"{self.tab}*size += {NUMERIC_TYPE_SIZES[self.protocol.string_size_type]} + {accessor}{var.name}_els_len[i{idx}];")
                        lines.append("}")
                    else:
                        lines.append(f"for ({self.get_native_list_size()} i{idx} = 0; i{idx} < {accessor}{var.name}_len; i{idx}++) {{")
                        self.indent_level += 1
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}[i{idx}].")
                        self.indent_level -= 1
                        if clines[0] == size_init:
                            clines = clines[1:]
                        clines.append(f"*size += {caccum};")
                        lines += [f"{self.tab}{l}" for l in clines]
                        lines.append("}")
                else:
                    if var.is_simple():
                        accum += self.protocol.get_size_of(var.vartype)
                    elif var.vartype == "string":
                        accum += NUMERIC_TYPE_SIZES[self.protocol.string_size_type]
                        lines.append(f"*size += {accessor}{var.name}_len;")
                    else:
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_default(self, members: list[Variable]):
        for var in members:
            if var.is_list:
                self.write_line(f".{var.name}_len = 0,")
                if var.vartype == "string":
                    self.write_line(f".{var.name}_els_len = NULL,")
                self.write_line(f".{var.name} = NULL,")
            elif var.vartype in self.base_defaults:
                self.write_line(f".{var.name} = {self.base_defaults[var.vartype]},")
            elif var.vartype == "string":
                self.write_line(f".{var.name}_len = 0,")
                self.write_line(f".{var.name} = (char*)\"\",")
            else:
                self.write_line(f".{var.name} = {{")
                self.indent_level += 1
                self.gen_default(self.protocol.structs[var.vartype].members)
                self.indent_level -= 1
                self.write_line("},")

    def destructor(self, var: Variable, accessor: str):
        if var.is_simple():
            return
        if var.is_list:
            if var.is_simple(True):
                self.write_line(f"free({accessor}{var.name});")
            else:
                idx = self.indent_level
                self.write_line(f"for ({self.get_native_list_size()} i{idx} = 0; i{idx} < {accessor}{var.name}_len; i{idx}++) {{")
                self.indent_level += 1
                inner = Variable(self.protocol, f"{var.name}[i{idx}]", var.vartype)
                self.destructor(inner, accessor)
                self.indent_level -= 1
                self.write_line("}")
                if var.vartype == "string":
                    self.write_line(f"free({accessor}{var.name}_els_len);")
                self.write_line(f"free({accessor}{var.name});")
        elif var.vartype == "string":
            self.write_line(f"free({accessor}{var.name});")
        else:
            [self.destructor(mem, f"{accessor}{var.name}.") for mem in self.protocol.structs[var.vartype].members]

    def gen_struct(self, sname: str, sdata: Struct):
        self.write_line("typedef struct {")
        self.indent_level += 1
        if sdata.is_message:
            self.write_line(f"{self.prefix}MessageType _mt;")
        for var in sdata.members:
            if var.is_list:
                self.write_line(f"{self.get_native_list_size()} {var.name}_len;")
                if var.vartype == "string":
                    self.write_line(f"{self.get_native_list_size()}* {var.name}_els_len;")
                if var.vartype in NUMERIC_TYPE_SIZES or var.vartype == "string":
                    self.write_line(f"{self.type_mapping[var.vartype]}* {var.name};")
                elif var.vartype in self.protocol.structs:
                    self.write_line(f"{self.prefix}{var.vartype}* {var.name};")
            elif var.vartype == "string":
                self.write_line(f"{self.get_native_string_size()} {var.name}_len;")
                self.write_line(f"{self.type_mapping[var.vartype]} {var.name};")
            elif var.vartype in NUMERIC_TYPE_SIZES:
                self.write_line(f"{self.type_mapping[var.vartype]} {var.name};")
            elif var.vartype in self.protocol.structs:
                self.write_line(f"{self.prefix}{var.vartype} {var.name};")
        self.indent_level -= 1
        self.write_line(f"}} {self.prefix}{sname};")

        if sdata.is_message:
            self.write_line(f"extern const {self.prefix}{sname} {self.prefix}{sname}_default;")
            self.write_line(f"const {self.prefix}{sname} {self.prefix}{sname}_default = {{")
            self.indent_level += 1
            self.write_line(f"._mt = {self.prefix}MessageType_{sname},")
            self.gen_default(sdata.members)
            self.indent_level -= 1
            self.write_line("};")
        self.write_line()
        if sdata.is_message:
            self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_WriteBytes({self.prefix}DataAccess* w, const {self.prefix}{sname}* src, bool tag);")
        else:
            self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_WriteBytes({self.prefix}DataAccess* w, const {self.prefix}{sname}* src);")
        self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_FromBytes({self.prefix}DataAccess* r, {self.prefix}{sname}* dst);")
        if sdata.is_message:
            self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_GetSizeInBytes(const {self.prefix}{sname}* m, size_t* size);")
            self.write_line(f"void {self.prefix}{sname}_Destroy({self.prefix}{sname} *m);")
        self.write_line()
        self.write_line()

    def gen_implementation(self, sname: str, sdata: Struct):
        if sdata.is_message:
            self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_GetSizeInBytes(const {self.prefix}{sname}* m, size_t* size) {{")
            self.indent_level += 1
            measure_lines, accumulator = self.gen_measurement(sdata, "m->")
            [self.write_line(s) for s in measure_lines]
            if accumulator > 0:
                self.write_line(f"*size += {accumulator};")
            self.write_line(f"return {self.prefix.upper()}ERR_OK;")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()

            self.write_line(f"void {self.prefix}{sname}_Destroy({self.prefix}{sname} *m) {{")
            self.indent_level += 1
            [self.destructor(mem, f"m->") for mem in sdata.members]
            self.write_line("free(m);")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()

        self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_FromBytes({self.prefix}DataAccess* r, {self.prefix}{sname}* dst) {{")
        self.indent_level += 1
        if sdata.is_message:
            self.write_line(f"dst->_mt = {self.prefix}MessageType_{sname};")
        if len(sdata.members) > 0:
            self.write_line(f"{self.prefix}err_t err;")
        [self.deserializer(mem, "dst->") for mem in sdata.members]
        self.write_line(f"return {self.prefix.upper()}ERR_OK;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if sdata.is_message:
            self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_WriteBytes({self.prefix}DataAccess* w, const {self.prefix}{sname}* src, bool tag) {{")
        else:
            self.write_line(f"{self.prefix}err_t {self.prefix}{sname}_WriteBytes({self.prefix}DataAccess* w, const {self.prefix}{sname}* src) {{")
        self.indent_level += 1
        self.write_line(f"{self.prefix}err_t err;")
        if sdata.is_message:
            self.write_line("if (tag) {")
            self.indent_level += 1
            self.write_line(f"err = {self.prefix}_WriteUInt8(w, (const uint8_t *)&(src->_mt));")
            self.err_check_return()
            self.indent_level -= 1
            self.write_line("}")
        [self.serializer(mem, "src->") for mem in sdata.members]
        self.write_line(f"return {self.prefix.upper()}ERR_OK;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def generate(self) -> str:
        self.output = []
        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}")
        self.write_line( "// <https://github.com/sjml/beschi>")
        self.write_line(f"// Do not edit directly.")
        self.write_line()

        if self.embed_protocol:
            self.write_line("/*")
            self.write_line("DATA PROTOCOL")
            self.write_line("-----------------")
            [self.write_line(f"{l}") for l in self.protocol.protocol_string.splitlines()]
            self.write_line("-----------------")
            self.write_line("END DATA PROTOCOL")
            self.write_line("*/")
            self.write_line()
            self.write_line()

        self.add_boilerplate(self.subs + [
            ("{# STRING_SIZE_TYPE #}", self.base_serializers[self.protocol.string_size_type]),
            ("{# STRING_SIZE_TYPE_LOWER #}", self.base_serializers[self.protocol.string_size_type].lower()),
            ("{# STRING_SIZE_TYPE_NATIVE #}", self.get_native_string_size()),
        ], 0)

        self.write_line("typedef enum {")
        self.indent_level += 1
        self.write_line(f"{self.prefix}MessageType___NullMessage = 0,")
        [self.write_line(f"{self.prefix}MessageType_{k} = {i+1}{',' if i < len(self.protocol.messages)-1 else ''}") for i, k in enumerate(self.protocol.messages.keys())]
        self.indent_level -= 1
        self.write_line(f"}} {self.prefix}MessageType;")
        self.write_line()
        self.write_line(f"{self.prefix}MessageType {self.prefix}GetMessageType(const void* m);")
        self.write_line(f"{self.prefix}err_t {self.prefix}GetSizeInBytes(const void* m, size_t* len);")
        self.write_line(f"{self.prefix}err_t {self.prefix}ProcessRawBytes({self.prefix}DataAccess* r, void*** msgListOut, size_t* len);")
        self.write_line(f"{self.prefix}err_t {self.prefix}DestroyMessageList(void** msgList, size_t len);")
        self.write_line()

        for sname, sdata in self.protocol.structs.items():
            self.gen_struct(sname, sdata)

        for mname, mdata in self.protocol.messages.items():
            self.gen_struct(mname, mdata)

        self.add_boilerplate(self.subs + [
            ("{# STRING_SIZE_TYPE #}", self.base_serializers[self.protocol.string_size_type]),
            ("{# STRING_SIZE_TYPE_LOWER #}", self.base_serializers[self.protocol.string_size_type].lower()),
        ], 1)

        self.write_line(f"{self.prefix}MessageType {self.prefix}GetMessageType(const void* m) {{")
        self.indent_level += 1
        self.write_line("const uint8_t* buffer = (const uint8_t*)m;")
        self.write_line("uint8_t msgType = buffer[0];")
        self.write_line(f"if (msgType > {len(self.protocol.messages)}) {{")
        self.indent_level += 1
        self.write_line(f"return {self.prefix}MessageType___NullMessage;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line(f"return ({self.prefix}MessageType)msgType;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line(f"{self.prefix}err_t {self.prefix}GetSizeInBytes(const void* m, size_t* len) {{")
        self.indent_level += 1
        self.write_line(f"{self.prefix}MessageType msgType = {self.prefix}GetMessageType(m);")
        self.write_line("switch (msgType) {")
        self.write_line(f"case {self.prefix}MessageType___NullMessage:")
        self.indent_level += 1
        self.write_line(f"return {self.prefix.upper()}ERR_INVALID_DATA;")
        self.write_line("break;")
        self.indent_level -= 1
        for msg_type in self.protocol.messages:
            self.write_line(f"case {self.prefix}MessageType_{msg_type}:")
            self.indent_level += 1
            self.write_line(f"return {self.prefix}{msg_type}_GetSizeInBytes((const {self.prefix}{msg_type}*)m, len);")
            self.write_line("break;")
            self.indent_level -= 1
        self.write_line("}")
        self.write_line(f"return {self.prefix.upper()}ERR_INVALID_DATA;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line(f"{self.prefix}err_t {self.prefix}ProcessRawBytes({self.prefix}DataAccess* r, void*** msgListDst, size_t* len) {{")
        self.indent_level += 1
        self.write_line(f"{self.prefix}err_t err = {self.prefix.upper()}ERR_OK;")
        self.write_line("size_t currCapacity = 8;")
        self.write_line("*msgListDst = (void**)malloc(sizeof(void*) * currCapacity);")
        self.write_line(f"if (*msgListDst == NULL) {{ return {self.prefix.upper()}ERR_ALLOCATION_FAILURE; }}")
        self.write_line("*len = 0;")
        self.write_line(f"while (!{self.prefix}IsFinished(r)) {{")
        self.indent_level += 1
        self.write_line("while (*len >= currCapacity) {")
        self.indent_level += 1
        self.write_line("currCapacity *= 2;")
        self.write_line("*msgListDst = (void**)realloc(*msgListDst, (sizeof(void*) * currCapacity));")
        self.write_line(f"if (*msgListDst == NULL) {{ return {self.prefix.upper()}ERR_ALLOCATION_FAILURE; }}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("uint8_t msgType;")
        self.write_line(f"{self.prefix}_ReadUInt8(r, &msgType);")
        self.err_check_return()
        self.write_line()
        self.write_line("void* out;")
        self.write_line("switch (msgType) {")
        for msg_type in self.protocol.messages:
            self.write_line(f"case {self.prefix}MessageType_{msg_type}:")
            self.indent_level += 1
            self.write_line(f"out = malloc(sizeof({self.prefix}{msg_type}));")
            self.write_line(f"if (out == NULL) {{ return {self.prefix.upper()}ERR_ALLOCATION_FAILURE; }}")
            self.write_line(f"err = {self.prefix}{msg_type}_FromBytes(r, ({self.prefix}{msg_type}*)out);")
            self.write_line("(*msgListDst)[*len] = out;")
            self.write_line("*len += 1;")
            self.err_check_return()
            self.write_line("break;")
            self.indent_level -= 1
        self.write_line("default:")
        self.indent_level += 1
        self.write_line(f"return {self.prefix.upper()}ERR_INVALID_DATA;")
        self.write_line("break;")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line(f"return {self.prefix.upper()}ERR_OK;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()
        self.write_line(f"{self.prefix}err_t {self.prefix}DestroyMessageList(void** msgList, size_t len) {{")
        self.indent_level += 1
        self.write_line("for (size_t i = 0; i < len; i++) {")
        self.indent_level += 1
        self.write_line(f"{self.prefix}MessageType msgType = {self.prefix}GetMessageType(msgList[i]);")
        self.write_line("switch (msgType) {")
        for msg_type in self.protocol.messages:
            self.write_line(f"case {self.prefix}MessageType_{msg_type}:")
            self.indent_level += 1
            self.write_line(f"{self.prefix}{msg_type}_Destroy(({self.prefix}{msg_type}*)msgList[i]);")
            self.write_line("break;")
            self.indent_level -= 1
        self.write_line(f"case {self.prefix}MessageType___NullMessage:")
        self.indent_level += 1
        self.write_line(f"return {self.prefix.upper()}ERR_INVALID_DATA;")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("free(msgList);")
        self.write_line(f"return {self.prefix.upper()}ERR_OK;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for sname, sdata in self.protocol.structs.items():
            self.gen_implementation(sname, sdata)

        for mname, mdata in self.protocol.messages.items():
            self.gen_implementation(mname, mdata)

        self.add_boilerplate(self.subs, 2)
        return "\n".join(self.output)



import argparse

from ..protocol import Protocol, Struct, Variable, NUMERIC_TYPE_SIZES
from ..writer import Writer, TextUtil
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "Go"


class GoWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".go"

    def get_additional_args(parser: argparse.ArgumentParser):
        group = parser.add_argument_group(LANGUAGE_NAME)
        group.add_argument("--go-no-rename", action="store_const", const=True, default=False, help="don't rename data members to Uppercase")

    def __init__(self, p: Protocol, extra_args: dict[str,any] = {}):
        rename = not extra_args["go_no_rename"]

        if rename:
            for _, s in p.structs.items():
                for var in s.members:
                    var.name = TextUtil.capitalize(var.name)
            for _, m in p.messages.items():
                for var in m.members:
                    var.name = TextUtil.capitalize(var.name)

        super().__init__(protocol=p, tab="\t")

        self.embed_protocol = extra_args["embed_protocol"]

        self.type_mapping["byte"] = "byte"
        self.type_mapping["bool"] = "bool"
        self.type_mapping["uint16"] = "uint16"
        self.type_mapping["int16"] = "int16"
        self.type_mapping["uint32"] = "uint32"
        self.type_mapping["int32"] = "int32"
        self.type_mapping["uint64"] = "uint64"
        self.type_mapping["int64"] = "int64"
        self.type_mapping["float"] = "float32"
        self.type_mapping["double"] = "float64"

    def deserializer(self, var: Variable, accessor: str):
        def err_panic():
            self.write_line("if err != nil {")
            self.write_line(f"{self.tab}panic(err)")
            self.write_line("}")

        if var.is_list:
            self.write_line(f"var {var.name}_Len {self.get_native_list_size()}")
            self.write_line(f"err = binary.Read(data, binary.LittleEndian, &{var.name}_Len)")
            err_panic()
            self.write_line(f"{accessor}.{var.name} = make([]{self.type_mapping[var.vartype]}, {var.name}_Len)")
            idx = self.indent_level
            self.write_line(f"for i{idx} := ({self.get_native_list_size()})(0); i{idx} < {var.name}_Len; i{idx}++ {{")
            self.indent_level += 1
            inner = Variable(self.protocol, f"{var.name}[i{idx}]", var.vartype)
            self.deserializer(inner, accessor)
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            self.write_line(f"err = readString(data, &{accessor}.{var.name})")
            err_panic()
        elif var.is_simple():
            self.write_line(f"err = binary.Read(data, binary.LittleEndian, &{accessor}.{var.name})")
            err_panic()
        else:
            self.write_line(f"{var.vartype}FromBytes(data, &{accessor}.{var.name})")

    def serializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"{var.name}_Len := ({self.get_native_list_size()})(len({accessor}.{var.name}))")
            self.write_line(f"binary.Write(data, binary.LittleEndian, {var.name}_Len)")
            idx = self.indent_level
            self.write_line(f"for i{idx} := ({self.get_native_list_size()})(0); i{idx} < {var.name}_Len; i{idx}++ {{")
            self.indent_level += 1
            inner = Variable(self.protocol, f"{var.name}[i{idx}]", var.vartype)
            self.serializer(inner, accessor)
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            self.write_line(f"writeString(data, &{accessor}.{var.name})")
        elif var.is_simple():
            self.write_line(f"binary.Write(data, binary.LittleEndian, &{accessor}.{var.name})")
        else:
            self.write_line(f"{accessor}.{var.name}.WriteBytes(data)")

    def gen_measurement(self, st: Struct, accessor: str = "") -> tuple[list[str], int]:
        lines: list[str] = []
        accum = 0

        if st.is_simple():
            lines.append(f"return {self.protocol.get_size_of(st.name)}")
        else:
            size_init = "size := 0"
            lines.append(size_init)

            for var in st.members:
                if var.is_list:
                    accum += NUMERIC_TYPE_SIZES[self.protocol.list_size_type]
                    if var.is_simple(True):
                        lines.append(f"size += len({accessor}{var.name}) * {self.protocol.get_size_of(var.vartype)}")
                    elif var.vartype == "string":
                        lines.append(f"for _, s := range {accessor}{var.name} {{")
                        lines.append(f"{self.tab}size += {NUMERIC_TYPE_SIZES[self.protocol.string_size_type]} + len(s)")
                        lines.append("}")
                    else:
                        lines.append(f"for _, el := range {accessor}{var.name} {{")
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"el.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        clines.append(f"size += {caccum}")
                        lines += [f"{self.tab}{l}" for l in clines]
                        lines.append("}")
                else:
                    if var.is_simple():
                        accum += self.protocol.get_size_of(var.vartype)
                    elif var.vartype == "string":
                        accum += NUMERIC_TYPE_SIZES[self.protocol.string_size_type]
                        lines.append(f"size += len({accessor}{var.name})")
                    else:
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_struct(self, sname: str, sdata: Struct):
        self.write_line(f"type {sname} struct {{")
        self.indent_level += 1
        for var in sdata.members:
            if var.is_list:
                self.write_line(f"{var.name} []{self.type_mapping[var.vartype]}")
            else:
                self.write_line(f"{var.name} {self.type_mapping[var.vartype]}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if sdata.is_message:
            self.write_line(f"func (output {sname}) GetMessageType() MessageType {{")
            self.indent_level += 1
            self.write_line(f"return {sname}Type")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()

            self.write_line(f"func (output {sname}) GetSizeInBytes() int {{")
            self.indent_level +=1
            measure_lines, accumulator = self.gen_measurement(sdata, "output.")
            [self.write_line(s) for s in measure_lines]
            if accumulator > 0:
                self.write_line(f"size += {accumulator}")
            if len(measure_lines) > 1:
                self.write_line(f"return size")
            self.indent_level -=1
            self.write_line("}")

            self.write_line(f"func {sdata.name}FromBytes (data io.Reader) (msg *{sname}) {{")
            self.indent_level += 1
            self.write_line("defer func() {")
            self.indent_level += 1
            self.write_line("if r := recover(); r != nil {")
            self.indent_level += 1
            self.write_line("msg = nil")
            self.indent_level -= 1
            self.write_line("}")
            self.indent_level -= 1
            self.write_line("}()")
            if len(sdata.members) > 0:
                self.write_line("var err error")
                self.write_line("err = nil")
            self.write_line(f"ret := {sname}{{}}")
            [self.deserializer(mem, "ret") for mem in sdata.members]
            if len(sdata.members) > 0:
                self.write_line("if err != nil {")
                self.indent_level += 1
                self.write_line(f"panic(err)")
                self.indent_level -= 1
                self.write_line("}")
            self.write_line()
            self.write_line("return &ret")
        else:
            self.write_line(f"func {sname}FromBytes (data io.Reader, input *{sname}) {{")
            self.indent_level += 1
            self.write_line("var err error")
            self.write_line("err = nil")
            [self.deserializer(mem, "input") for mem in sdata.members]
            self.write_line("if err != nil {")
            self.indent_level += 1
            self.write_line(f"panic(err)")
            self.indent_level -= 1
            self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if sdata.is_message:
            self.write_line(f"func (output {sname}) WriteBytes (data io.Writer, tag bool) {{")
            self.indent_level += 1
            self.write_line("if tag {")
            self.indent_level += 1
            self.write_line(f"binary.Write(data, binary.LittleEndian, {sname}Type)")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line(f"func (output {sname}) WriteBytes (data io.Writer) {{")
            self.indent_level += 1
        [self.serializer(mem, "output") for mem in sdata.members]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def generate(self) -> str:
        self.output = []

        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}")
        self.write_line( "// <https://github.com/sjml/beschi>")
        self.write_line(f"// Do not edit directly.")
        self.write_line()

        if self.embed_protocol:
            self.write_line("/*")
            self.write_line("DATA PROTOCOL")
            self.write_line("-----------------")
            [self.write_line(f"{l}") for l in self.protocol.protocol_string.splitlines()]
            self.write_line("-----------------")
            self.write_line("END DATA PROTOCOL")
            self.write_line("*/")
            self.write_line()
            self.write_line()


        subs = [("{# STRING_SIZE_TYPE #}", self.get_native_string_size())]
        if self.protocol.namespace:
            subs.append(("Beschi", self.protocol.namespace))

        self.add_boilerplate(subs)
        self.write_line()

        self.write_line("type MessageType byte")
        self.write_line("const (")
        self.indent_level += 1
        [self.write_line(f"{msg_name}Type MessageType = {i+1}") for i, msg_name in enumerate(self.protocol.messages)]
        self.indent_level -= 1
        self.write_line(")")
        self.write_line()

        self.write_line("type Message interface {")
        self.indent_level += 1
        self.write_line("GetMessageType() MessageType")
        self.write_line("WriteBytes(data io.Writer, tag bool)")
        self.write_line("GetSizeInBytes() int")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("func ProcessRawBytes (data io.Reader) []Message {")
        self.indent_level += 1
        self.write_line("var msgList []Message")
        self.write_line("var err error")
        self.write_line("for err != io.EOF {")
        self.indent_level += 1
        self.write_line("var msgType MessageType")
        self.write_line("err = binary.Read(data, binary.LittleEndian, &msgType)")
        self.write_line("if err == io.EOF {")
        self.indent_level += 1
        self.write_line("break")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("switch msgType {")
        for msg_type in self.protocol.messages:
            self.write_line(f"case {msg_type}Type:")
            self.indent_level += 1
            self.write_line(f"msgList = append(msgList, {msg_type}FromBytes(data))")
            self.indent_level -= 1
        self.write_line("default:")
        self.indent_level += 1
        self.write_line("msgList = append(msgList, nil)")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("if msgList[len(msgList)-1] == nil {")
        self.indent_level += 1
        self.write_line("break")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("return msgList")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for sname, sdata in self.protocol.structs.items():
            self.gen_struct(sname, sdata)

        for mname, mdata in self.protocol.messages.items():
            self.gen_struct(mname, mdata)

        self.write_line()
        assert self.indent_level == 0

        return "\n".join(self.output)



from ..protocol import Protocol, Struct, Variable, NUMERIC_TYPE_SIZES
from ..writer import Writer, TextUtil
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "CSharp"


class CSharpWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".cs"

    def __init__(self, p: Protocol, extra_args: dict[str,any] = {}):
        super().__init__(protocol=p, tab="    ")

        self.embed_protocol = extra_args["embed_protocol"]

        self.type_mapping["byte"] = "byte"
        self.type_mapping["bool"] = "bool"
        self.type_mapping["uint16"] = "ushort"
        self.type_mapping["int16"] = "short"
        self.type_mapping["uint32"] = "uint"
        self.type_mapping["int32"] = "int"
        self.type_mapping["uint64"] = "ulong"
        self.type_mapping["int64"] = "long"
        self.type_mapping["float"] = "float"
        self.type_mapping["double"] = "double"

        self.base_deserializers: dict[str,str] = {
            "byte": "ReadByte",
            "bool": "ReadBoolean",
            "uint16": "ReadUInt16",
            "int16": "ReadInt16",
            "uint32": "ReadUInt32",
            "int32": "ReadInt32",
            "uint64": "ReadUInt64",
            "int64": "ReadInt64",
            "float": "ReadSingle",
            "double": "ReadDouble",
        }

    def deserializer(self, var: Variable, accessor: str):
        var_clean = TextUtil.replace(var.name, [("[", "_"), ("]", "_"), (" ", "_")])
        if var.is_list:
            self.write_line(f"{self.get_native_list_size()} {var_clean}_Length = br.{self.base_deserializers[self.protocol.list_size_type]}();")
            self.write_line(f"{accessor}{var.name} = new List<{self.type_mapping[var.vartype]}>();")
            idx = self.indent_level
            self.write_line(f"for (int i{idx} = 0; i{idx} < {var_clean}_Length; i{idx}++)")
            self.write_line("{")
            self.indent_level += 1
            inner = Variable(self.protocol, f"{self.type_mapping[var.vartype]} el", var.vartype)
            self.deserializer(inner, "")
            self.write_line(f"{accessor}{var.name}.Add(el);")
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            self.write_line(f"{self.get_native_string_size()} {var_clean}_Length = br.{self.base_deserializers[self.protocol.string_size_type]}();")
            self.write_line(f"byte[] {var_clean}_Buffer = br.ReadBytes((int){var_clean}_Length);")
            self.write_line(f"{accessor}{var.name} = System.Text.Encoding.UTF8.GetString({var_clean}_Buffer);")
        elif var.vartype in self.base_deserializers:
            self.write_line(f"{accessor}{var.name} = br.{self.base_deserializers[var.vartype]}();")
        else:
            self.write_line(f"{accessor}{var.name} = {var.vartype}.FromBytes(br);")

    def serializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"bw.Write(({self.get_native_list_size()}){accessor}{var.name}.Count);")
            self.write_line(f"foreach ({self.type_mapping[var.vartype]} el in {accessor}{var.name})")
            self.write_line("{")
            self.indent_level += 1
            inner = Variable(self.protocol, "el", var.vartype)
            self.serializer(inner, "")
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype == "string":
            self.write_line(f"byte[] {var.name}_Buffer = System.Text.Encoding.UTF8.GetBytes({accessor}{var.name});")
            self.write_line(f"bw.Write(({self.get_native_string_size()}){var.name}_Buffer.Length);")
            self.write_line(f"bw.Write({var.name}_Buffer);")
        elif var.vartype in NUMERIC_TYPE_SIZES:
            self.write_line(f"bw.Write({accessor}{var.name});")
        else:
            self.write_line(f"{accessor}{var.name}.WriteBytes(bw);")

    def gen_measurement(self, st: Struct, accessor: str = "") -> tuple[list[str], int]:
        lines: list[str] = []
        accum = 0

        if st.is_simple():
            lines.append(f"return {self.protocol.get_size_of(st.name)};")
        else:
            size_init = "int size = 0;"
            lines.append(size_init)

            for var in st.members:
                if var.is_list:
                    accum += NUMERIC_TYPE_SIZES[self.protocol.list_size_type]
                    if var.is_simple(True):
                        lines.append(f"size += {accessor}{var.name}.Count * {self.protocol.get_size_of(var.vartype)};")
                    elif var.vartype == "string":
                        lines.append(f"foreach (string s in {accessor}{var.name})")
                        lines.append("{")
                        lines.append(f"{self.tab}size += {NUMERIC_TYPE_SIZES[self.protocol.string_size_type]} + System.Text.Encoding.UTF8.GetBytes(s).Length;")
                        lines.append("}")
                    else:
                        idx = self.indent_level
                        lines.append(f"foreach ({self.type_mapping[var.vartype]} el{idx} in {accessor}{var.name})")
                        lines.append("{")
                        self.indent_level += 1
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"el{idx}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        clines.append(f"size += {caccum};")
                        lines += [f"{self.tab}{l}" for l in clines]
                        self.indent_level -= 1
                        lines.append("}")
                else:
                    if var.is_simple():
                        accum += self.protocol.get_size_of(var.vartype)
                    elif var.vartype == "string":
                        accum += NUMERIC_TYPE_SIZES[self.protocol.string_size_type]
                        lines.append(f"size += {accessor}{var.name}.Length;")
                    else:
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_struct(self, sname: str, sdata: Struct):
        if sdata.is_message:
            self.write_line(f"public class {sname} : Message")
        else:
            self.write_line(f"public class {sname}")
        self.write_line("{")
        self.indent_level += 1

        for var in sdata.members:
            if var.is_list:
                self.write_line(f"public List<{self.type_mapping[var.vartype]}> {var.name} = new List<{self.type_mapping[var.vartype]}>();")
            else:
                default = None
                if var.vartype == "string":
                    default = '""'
                elif var.vartype in self.protocol.structs:
                    default = f"new {var.vartype}()"
                self.write_line(f"public {self.type_mapping[var.vartype]} {var.name}{f' = {default}' if default else ''};")

        if sdata.is_message:
            self.write_line()
            self.write_line(f"public override MessageType GetMessageType() {{ return MessageType.{sname}Type; }}")
            self.write_line()
            self.write_line("public override int GetSizeInBytes()")
            self.write_line("{")
            self.indent_level +=1
            measure_lines, accumulator = self.gen_measurement(sdata, "this.")
            [self.write_line(s) for s in measure_lines]
            if accumulator > 0:
                self.write_line(f"size += {accumulator};")
            if len(measure_lines) > 1:
                self.write_line(f"return size;")
            self.indent_level -=1
            self.write_line("}")
        self.write_line()

        self.write_line(f"public static {sname} FromBytes(BinaryReader br)")
        self.write_line("{")
        self.indent_level += 1
        if sdata.is_message:
            self.write_line("try")
            self.write_line("{")
            self.indent_level += 1
        self.write_line(f"{sname} n{sname} = new {sname}();")
        [self.deserializer(mem, f"n{sname}.") for mem in sdata.members]
        self.write_line(f"return n{sname};")
        if sdata.is_message:
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("catch (System.IO.EndOfStreamException)")
            self.write_line("{")
            self.indent_level += 1
            self.write_line("return null;")
            self.indent_level -= 1
            self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")

        if sdata.is_message:
            self.write_line(f"public override void WriteBytes(BinaryWriter bw, bool tag)")
            self.write_line("{")
            self.indent_level += 1
            self.write_line("if (tag)")
            self.write_line("{")
            self.indent_level += 1
            self.write_line(f"bw.Write((byte)MessageType.{sname}Type);")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line(f"public void WriteBytes(BinaryWriter bw)")
            self.write_line("{")
            self.indent_level += 1
        [self.serializer(mem, "this.") for mem in sdata.members]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def generate(self) -> str:
        self.output = []

        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}")
        self.write_line( "// <https://github.com/sjml/beschi>")
        self.write_line(f"// Do not edit directly.")
        self.write_line()

        if self.embed_protocol:
            self.write_line("/*")
            self.write_line("DATA PROTOCOL")
            self.write_line("-----------------")
            [self.write_line(f"{l}") for l in self.protocol.protocol_string.splitlines()]
            self.write_line("-----------------")
            self.write_line("END DATA PROTOCOL")
            self.write_line("*/")
            self.write_line()
            self.write_line()

        self.write_line("using System;")
        self.write_line("using System.IO;")
        self.write_line("using System.Text;")
        self.write_line("using System.Collections.Generic;")
        self.write_line()

        if self.protocol.namespace:
            self.write_line(f"namespace {self.protocol.namespace}")
            self.write_line("{")
            self.indent_level += 1

        self.write_line("public enum MessageType")
        self.write_line("{")
        self.indent_level += 1
        [self.write_line(f"{k}Type = {i+1},") for i, k in enumerate(self.protocol.messages)]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("public abstract class Message {")
        self.indent_level += 1
        self.write_line("abstract public MessageType GetMessageType();")
        self.write_line("abstract public void WriteBytes(BinaryWriter bw, bool tag);")
        self.write_line("abstract public int GetSizeInBytes();")
        self.write_line()

        self.write_line("public static Message[] ProcessRawBytes(BinaryReader br)")
        self.write_line("{")
        self.indent_level += 1
        self.write_line("List<Message> msgList = new List<Message>();")
        self.write_line("while (br.BaseStream.Position < br.BaseStream.Length)")
        self.write_line("{")
        self.indent_level += 1
        self.write_line("byte msgType = br.ReadByte();")
        self.write_line("switch (msgType)")
        self.write_line("{")
        self.indent_level += 1
        for msg_type in self.protocol.messages:
            self.write_line(f"case (byte)MessageType.{msg_type}Type:")
            self.indent_level += 1
            self.write_line(f"msgList.Add({msg_type}.FromBytes(br));")
            self.write_line("break;")
            self.indent_level -= 1
        self.write_line("default:")
        self.indent_level += 1
        self.write_line("msgList.Add(null);")
        self.write_line("break;")
        self.indent_level -= 1
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("if (msgList[msgList.Count-1] == null) {")
        self.indent_level += 1
        self.write_line("break;")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("return msgList.ToArray();")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for sname, sdata in self.protocol.structs.items():
            self.gen_struct(sname, sdata)

        for mname, mdata in self.protocol.messages.items():
            self.gen_struct(mname, mdata)

        if self.protocol.namespace:
            self.indent_level -= 1
            self.write_line("}")

        self.write_line()
        assert self.indent_level == 0

        return "\n".join(self.output)



import argparse

from ..protocol import Protocol, Struct, Variable, NUMERIC_TYPE_SIZES
from ..writer import Writer, TextUtil
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "Rust"


class RustWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".rs"

    def get_additional_args(parser: argparse.ArgumentParser):
        group = parser.add_argument_group(LANGUAGE_NAME)
        group.add_argument("--rust-no-rename", action="store_const", const=True, default=False, help="don't rename data members to snake_case")

    def __init__(self, p: Protocol, extra_args: dict[str,any] = {}):
        rename = not extra_args["rust_no_rename"]

        if rename:
            for _, s in p.structs.items():
                for var in s.members:
                    var.name = TextUtil.convert_to_lower_snake_case(var.name)
            for _, m in p.messages.items():
                for var in m.members:
                    var.name = TextUtil.convert_to_lower_snake_case(var.name)

        super().__init__(protocol=p, tab="    ")

        self.embed_protocol = extra_args["embed_protocol"]

        self.type_mapping["byte"] = "u8"
        self.type_mapping["bool"] = "bool"
        self.type_mapping["uint16"] = "u16"
        self.type_mapping["int16"] = "i16"
        self.type_mapping["uint32"] = "u32"
        self.type_mapping["int32"] = "i32"
        self.type_mapping["uint64"] = "u64"
        self.type_mapping["int64"] = "i64"
        self.type_mapping["float"] = "f32"
        self.type_mapping["double"] = "f64"
        self.type_mapping["string"] = "String"

    def deserializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"let {var.name}_len = reader.read_{self.get_native_list_size()}()?;")
            if var.vartype in self.type_mapping:
                self.write_line(f"let mut {var.name}: Vec<{self.type_mapping[var.vartype]}> = Vec::new();")
            else:
                self.write_line(f"let mut {var.name}: Vec<{var.vartype}> = Vec::new();")
            self.write_line(f"for _ in 0..{var.name}_len {{")
            self.indent_level += 1
            inner = Variable(self.protocol, "el", var.vartype)
            self.deserializer(inner, "");
            self.write_line(f"{var.name}.push(el);")
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype in NUMERIC_TYPE_SIZES:
            if var.vartype == "byte":
                self.write_line(f"let {var.name} = reader.take_byte()?;")
            elif var.vartype == "bool":
                self.write_line(f"let {var.name} = reader.take_byte()? > 0;")
            else:
                self.write_line(f"let {var.name} = reader.read_{self.type_mapping[var.vartype]}()?;")
        elif var.vartype == "string":
            self.write_line(f"let {var.name} = reader.read_string()?;")
        else:
            self.write_line(f"let {var.name} = {var.vartype}::from_bytes(reader)?;")

    def serializer(self, var: Variable, accessor: str):
        if var.is_list:
            self.write_line(f"writer.extend(({accessor}{var.name}.len() as {self.get_native_list_size()}).to_le_bytes());")
            self.write_line(f"for el in &{accessor}{var.name} {{")
            self.indent_level += 1
            inner = Variable(self.protocol, "el", var.vartype)
            self.serializer(inner, "")
            self.indent_level -= 1
            self.write_line("}")
        elif var.vartype in NUMERIC_TYPE_SIZES:
            if var.vartype == "byte":
                self.write_line(f"writer.push({accessor}{var.name});")
            elif var.vartype == "bool":
                self.write_line(f"writer.push(if {accessor}{var.name} {{1_u8}} else {{0_u8}});")
            else:
                self.write_line(f"writer.extend({accessor}{var.name}.to_le_bytes());")
        elif var.vartype == "string":
            self.write_line(f"writer.extend(({accessor}{var.name}.len() as {self.get_native_string_size()}).to_le_bytes());")
            self.write_line(f"writer.extend({accessor}{var.name}.as_bytes());")
        else:
            self.write_line(f"{accessor}{var.name}.write_bytes(writer);")

    def gen_measurement(self, st: Struct, accessor: str = "") -> tuple[list[str], int]:
        lines: list[str] = []
        accum = 0

        if st.is_simple():
            lines.append(f"{self.protocol.get_size_of(st.name)}")
        else:
            size_init = "let mut size: u32 = 0;"
            lines.append(size_init)

            for var in st.members:
                if var.is_list:
                    accum += NUMERIC_TYPE_SIZES[self.protocol.list_size_type]
                    if var.is_simple(True):
                        lines.append(f"size += ({accessor}{var.name}.len() as u32) * {self.protocol.get_size_of(var.vartype)};")
                    elif var.vartype == "string":
                        lines.append(f"for s in &{accessor}{var.name} {{")
                        lines.append(f"{self.tab}size += {NUMERIC_TYPE_SIZES[self.protocol.string_size_type]} + (s.len() as u32);")
                        lines.append("}")
                    else:
                        lines.append(f"for el in &{accessor}{var.name} {{")
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"el.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        clines.append(f"size += {caccum};")
                        lines += [f"{self.tab}{l}" for l in clines]
                        lines.append("}")
                else:
                    if var.is_simple():
                        accum += self.protocol.get_size_of(var.vartype)
                    elif var.vartype == "string":
                        accum += NUMERIC_TYPE_SIZES[self.protocol.string_size_type]
                        lines.append(f"size += {accessor}{var.name}.len() as u32;")
                    else:
                        clines, caccum = self.gen_measurement(self.protocol.structs[var.vartype], f"{accessor}{var.name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_struct(self, sname: str, sdata: Struct):
        self.write_line("#[derive(Default)]")
        self.write_line(f"pub struct {sname} {{")
        self.indent_level += 1
        for var in sdata.members:
            if var.is_list:
                if var.vartype in self.type_mapping:
                    self.write_line(f"pub {var.name}: Vec<{self.type_mapping[var.vartype]}>,")
                else:
                    self.write_line(f"pub {var.name}: Vec<{var.vartype}>,")
            elif var.vartype in self.type_mapping:
                self.write_line(f"pub {var.name}: {self.type_mapping[var.vartype]},")
            else:
                self.write_line(f"{var.name}: {var.vartype},")
        self.indent_level -= 1
        self.write_line("}")

        self.write_line(f"impl {sname} {{")
        self.indent_level += 1

        self.write_line("pub fn get_size_in_bytes(&self) -> u32 {")
        self.indent_level += 1
        measure_lines, accumulator = self.gen_measurement(sdata, "self.")
        [self.write_line(s) for s in measure_lines]
        if accumulator > 0:
            self.write_line(f"size += {accumulator};")
        if len(measure_lines) > 1:
            self.write_line("size")
        self.indent_level -= 1
        self.write_line("}")

        self.write_line(f"pub fn from_bytes({'_' if len(sdata.members) == 0 else ''}reader: &mut BufferReader) -> Result<{sname}, {self.prefix}Error> {{")
        self.indent_level += 1
        [self.deserializer(mem, "") for mem in sdata.members]
        varnames = [mem.name for mem in sdata.members]
        self.write_line(f"Ok({sname} {{{', '.join(varnames)}}})")
        self.indent_level -= 1
        self.write_line("}")

        if sdata.is_message:
            self.write_line("pub fn write_bytes(&self, writer: &mut Vec<u8>, tag: bool) {")
            self.indent_level += 1
            self.write_line("if tag {")
            self.indent_level += 1
            msg_type_id = list(self.protocol.messages.keys()).index(sname) + 1
            self.write_line(f"writer.push({msg_type_id}_u8);")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line("pub fn write_bytes(&self, writer: &mut Vec<u8>) {")
            self.indent_level += 1
        [self.serializer(mem, "self.") for mem in sdata.members]
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1

        self.write_line("}")
        self.write_line()

    def generate(self) -> str:
        self.output = []
        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}")
        self.write_line( "// <https://github.com/sjml/beschi>")
        self.write_line(f"// Do not edit directly.")
        self.write_line()

        if self.embed_protocol:
            self.write_line("/*")
            self.write_line("DATA PROTOCOL")
            self.write_line("-----------------")
            [self.write_line(f"{l}") for l in self.protocol.protocol_string.splitlines()]
            self.write_line("-----------------")
            self.write_line("END DATA PROTOCOL")
            self.write_line("*/")
            self.write_line()
            self.write_line()

        subs = [("{# STRING_SIZE_TYPE #}", self.get_native_string_size())]
        self.prefix = "Beschi"
        if self.protocol.namespace != None:
            subs.append(("Beschi", self.protocol.namespace))
            self.prefix = self.protocol.namespace
        self.add_boilerplate(substitutions=subs)

        self.write_line("pub enum Message {")
        self.indent_level += 1
        for mname in self.protocol.messages:
           self.write_line(f"{mname}({mname}),")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line(f"pub fn process_raw_bytes(reader: &mut BufferReader) -> Result<Vec<Message>, {self.prefix}Error> {{")
        self.indent_level += 1
        self.write_line("let mut msg_list: Vec<Message> = Vec::new();")
        self.write_line("while !reader.is_finished() {")
        self.indent_level += 1
        self.write_line("let msg_type = reader.take_byte()?;")
        self.write_line("match msg_type {")
        self.indent_level += 1
        for i, k in enumerate(self.protocol.messages.keys()):
            self.write_line(f"{i+1} => msg_list.push(Message::{k}({k}::from_bytes(reader)?)),")
        self.write_line(f"_ => return Err({self.prefix}Error::InvalidData),")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("Ok(msg_list)")
        self.indent_level -= 1
        self.write_line("}")

        for sname, sdata in self.protocol.structs.items():
            self.gen_struct(sname, sdata)

        for mname, mdata in self.protocol.messages.items():
            self.gen_struct(mname, mdata)

        return "\n".join(self.output)



import os
import importlib

from ..writer import Writer


all_writers: dict[str, Writer] = {}
experimental_writers: dict[str, Writer] = {}

_writer_dir = os.path.abspath(os.path.dirname(__file__))
for entry in os.listdir(_writer_dir):
    if not entry.endswith(".py"): continue
    if entry == "__init__.py": continue

    writer_module = importlib.import_module(f"beschi.writers.{os.path.splitext(entry)[0]}")
    writer_class = getattr(writer_module, f"{writer_module.LANGUAGE_NAME}Writer")
    if not writer_class.in_progress:
        all_writers[writer_module.LANGUAGE_NAME.lower()] = writer_class
    else:
        experimental_writers[writer_module.LANGUAGE_NAME.lower()] = writer_class




import os
import re
from setuptools import setup, find_packages

from beschi import LIB_VERSION, LIB_NAME

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), encoding="utf-8", mode="r") as f:
    LONGDESC = f.read()
    SHORTDESC = re.search(r'bit-packing and unpacking code generator for [^\.]*', LONGDESC).group()

setup(
    name=LIB_NAME.lower(),
    version=LIB_VERSION,
    description=SHORTDESC,
    long_description=LONGDESC,
    long_description_content_type='text/markdown',
    url='https://github.com/sjml/beschi',
    author='Shane Liesegang',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',
    ],
    include_package_data=True,
    package_data={'beschi': ['writers/boilerplate/*']},
    packages=find_packages(
        include=['beschi', 'beschi.writers'],
        exclude=['test'], # seems to include test_util no matter what. ¯\_(ツ)_/¯
    ),
    install_requires=[
        'toml',
    ],
    extras_require={
        'dev': [
            'pytest',
        ],
        'publish': [
            'build',
            'twine'
        ]
    },
    entry_points={
        'console_scripts': [
            'beschi = beschi.cli:main'
        ]
    }
)



import sys
import argparse

from .protocol import Protocol
from .writers import all_writers, experimental_writers
from . import LIB_NAME, LIB_VERSION

VERSION_STRING = f"{LIB_NAME} v{LIB_VERSION}"

def main():
    writers = [w.lower() for w in all_writers]
    exp_writers = [w.lower() for w in experimental_writers]

    argparser = argparse.ArgumentParser(description="Generate the code for reading/writing messages from a given protocol.")
    argparser.add_argument("--version", "-v", action="store_const", const=True, default=False, help="print the version and exit")
    argparser.add_argument("--lang", "-l", type=str, help=f"language to generate ({' '.join(sorted(writers))})")
    argparser.add_argument("--output", "-o", type=str, default=None, help="path to output file; if omitted, will output to stdout")
    argparser.add_argument("--protocol", "-p", type=str, help="path to the protocol TOML file")

    additional_arguments = []
    extra_parser = argparser.add_argument_group("Extra")
    extra_parser.add_argument("--embed-protocol", action="store_const", const=True, default=False, help="embed the text of the protocol as a comment at the top of the generated file")
    additional_arguments.append("embed_protocol")

    for _, wclass in all_writers.items():
        wclass.get_additional_args(argparser)

    argparser.epilog = "For more usage information, visit the development site: https://github.com/sjml/beschi"

    args = argparser.parse_args()

    if args.version:
        print(VERSION_STRING)
        sys.exit(0)

    if args.protocol == None:
        sys.stderr.write("ERROR: Missing protocol. Specify a file with '--protocol FILENAME'.\n")
        sys.exit(1)
    if args.lang == None:
        sys.stderr.write(f"ERROR: Missing language. Specify a language for output with '--lang {{{'|'.join(writers)}}}'.\n")
        sys.exit(1)

    try:
        protocol = Protocol(args.protocol)
    except FileNotFoundError:
        sys.stderr.write(f"ERROR: No such file: {args.protocol}\n")
        sys.exit(1)
    except KeyError as ke:
        sys.stderr.write(f"ERROR: Invalid protocol file '{args.protocol}' is missing {ke}.\n")
        sys.exit(1)
    except RecursionError as rec:
        sys.stderr.write(f"ERROR: Invalid protocol file '{args.protocol}' has infinite loop: {rec}\n")
        sys.exit(1)
    except Exception as e:
        sys.stderr.write(f"ERROR: Invalid protocol file '{args.protocol}'.\n")
        sys.stderr.write(f"\t{e}\n")
        sys.exit(1)

    if args.lang not in writers and args.lang not in exp_writers:
        sys.stderr.write("\n".join([
            "ERROR: invalid language. Valid writers are:",
            "\t" + ", ".join(writers),
            ""
        ]))
        sys.exit(1)

    try:
        writer_class = all_writers[args.lang]
    except KeyError:
        writer_class = experimental_writers[args.lang]

    extra_args = {flag: val for flag, val in args._get_kwargs() if flag.startswith(args.lang) or flag in additional_arguments}
    writer = writer_class(protocol, extra_args=extra_args)
    try:
        output = writer.generate()
    except NotImplementedError as nie:
        sys.stderr.write(f"{nie}\n")
        sys.exit(1)

    if args.output == None:
        print(output)
    else:
        # if there's any problem with creating/writing to this file, just let
        #    the regular exceptions bubble out
        with open(args.output, "w") as output_file:
            output_file.write(output)
